/*
    (C) Gustav Hartvigsson, 2013.
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "baseobject.h"

typedef struct _BaseObjectClass {
  // BaseObjectInstance * (* initize)(* BaseObjectInstance);
  void (* deinitize)(* BaseObjectInstance);
  int (* ref)(* BaseObjectInstance);
  int (* unref)(* BaseObjectInstance);
  int (* get_refcount)(* BaseObectInstance);
  char * (* to_string)(* BaseObjectInstance);
};

typedef struct _BaseObjectInstance {
  BaseObjectClass * base_class;
  unsigned int refcount = 0;
};

/* ---------------------------
 * Concrete method definitions
 * These are implemented towards the end of this file.
 * ---------------------------
 */

// BaseObjectInstance * method_base_init (BaseObjectInstance * self);

void method_base_deinit (BaseObjectInstance * self);

int method_base_ref (BaseObjectInstance * self);

int method_base_unref (BaseObjectInstance * self);

int method_base_get_refcount (BaseObjectInstance * self);

char * method_base_to_string (BaseObjectInstance * self);

/* -----------------
 * Helper functions...
 * -----------------
 */
/**
 * This function is used to set the method to initize a new instance of an
 * object.
 *
 * This has no use, rely...(?)
 */ /*
void base_object_set_init_method (BaseObjectInstance * self, BaseObjectInstance * (* method)(* BaseObjectInstance)) {
  BaseObjectClass * klass = base_object_get_class (self);
  klass->initize = method;
}
*/

/**
 * This function is used to set the method to deinitize an object.
 * 
 * set it to a method that deinitize your object.
 */
void base_object_set_deinit_method (BaseObjectInstance * self, void (* method)(* BaseObjectInstance)) {
  BaseObjectClass * klass = base_object_get_class (self);
  klass->deinitize = method;
} 

/**
 * This fuction is used to set the ref method.
 * 
 * DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.
 */
void base_object_set_ref_method (BaseObjectInstance * self, int (* method)(* BaseObjectInstance)) {
  BaseObjectClass * klass = base_object_get_class (self);
  klass->ref = method;
}

/**
 * This fuction is used to set the unref method.
 * 
 * DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.
 */
void base_object_set_unref_method (BaseObjectInstance * self, int (* method)(* BaseObjectInstance)) {
  BaseObjectClass * klass = base_object_get_class (self);
  klass->unref = method;
}

/**
 * This fuction is used to set the get_refcount method.
 * 
 * DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.
 */
void base_object_set_unref_method (BaseObjectInstance * self, int (* method)(* BaseObjectInstance)) {
  BaseObjectClass * klass = base_object_get_class (self);
  klass->get_refcount = method;
}

/**
 * This fuction is used to set the to_string method.
 */
void base_object_set_to_string_method (BaseObjectInstance * self, char * (* method)(* BaseObjectInstance)) {
  BaseObjectClass * klass = base_object_get_class (self);
  klass->to_string = method;
}

/* ----------------------
 * Base object functions.
 * ----------------------
 */

/**
 * This function initizes an intance of the BaseObject, it also sets the methods
 * to be used with the object and sets the refrence count to one.
 */
void base_object_initize (BaseObjectInstance * self) {
  self->refcount = 1;
  // base_object_set_init_method (self, method_base_init);
  base_object_set_deintit_method (self, method_base_deinit);
  base_object_set_ref_method (self, method_base_ref);
  base_object_set_unref_method (self, method_base_unref);
  base_object_set_get_ref_count_method (self, method_base_get_refcount);
  base_object_set_to_string_method (self, method_base_to_string);
}

/**
 * This function creates a new base object.
 */
BaseObjectInstance * base_object_new () {
  BaseObjectInstance * self = malloc (sizeof (BaseObjectInstance) + 1);
  //allocate the class definition of the object.
  self->base_class = malloc (class_size + 1);
  //initize it.
  base_object_initize (self, sizeof(BaseObjectClass) + 1);
  return self;
}

/**
 * This function deinitizes/frees an object even if it is still referenced.
 * This is usualy a bad idea, use base_object_unref instead.
 */
void base_object_free (BaseObjectInstance * self) {
  BaseObjectClass * klass = base_object_get_class (self);
  klass->deinitize (self);
}

/**
 * This function gets the class (which hold the object methods).
 */
BaseObjectClass * base_object_get_class (BaseObjectInstance * self) {
  return self->base_class;
}

/**
 * This function sets the instance class of an object.
 */
void base_object_set_class (BaseObjectInstance * self, BaseObjectClass * klass) {
  self->base_class = klass;
}

/**
 * This function is used to decrese the reference count of an object.
 * When an object reaches zero, it will deinitize the object using the objects
 * deititize method.
 * 
 * It returns the current reference count.
 */
int base_object_unref (BaseObjectInstance * self) {
  return self->base_class->unref (self);
}

/**
 * This function is used to increse the reference count of an object.
 *
 * Returns the current reference count.
 */
int base_object_ref (BaseObjectInstance * self) {
  return self->base_class->ref (self);
}

/**
 * This function returns the current reference count without chaning it.
 */
int base_object_get_refcount (BaseObjectInstance * self) {
  return self->base_class->get_refcount (self);
}

/**
 * This function returns a textual (string) that represesnts the object.
 * The method can be set using base_object_set_to_string_method.
 *
 * Note: The string that is returned must be freed.
 */
char * base_object_to_string (BaseObjectInstance * self) {
  return self->base_class->to_string (self);
}

/* -----------------
 * Contrete methods.
 * -----------------
 */
 
 
// BaseObjectInstance * method_base_init (BaseObjectInstance * self) { }

void method_base_deinit (BaseObjectInstance * self) {
  free (self->base_class);
  free (self);
}

int method_base_ref (BaseObjectInstance * self) {
  self->refcount = self->refcount + 1;
  return self->refcount;
}

int method_base_unref (BaseObjectInstance * self) {
  self->refcount = self->refcount - 1;
  if (self->refcount <= 0) {
    self->base_class->deinit (self);
    return 0;
  }
  return self->refcount;
}

int method_base_get_refcount (BaseObjectInstance * self) {
  return self->refcount;
}

char * method_base_to_string (BaseObjectInstance * self) {
  char * ret_string = malloc (sizeof (char) * 32); // I am bad with maths..
  snprintf (ret_string, "References: %d ", self->base_class->get_refcount(self));
  return ret_string;
}


