/*
    (C) Gustav Hartvigsson, 2013.
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __H_BASE_OBJECT__
#define __H_BASE_OBJECT__

/* ---------------------------------------------------
 * This is just a test of creating a small typesystem.
 * It will include refrence counting.
 * ---------------------------------------------------
 */

/**
 * The class holds all the "virtual" functions/methods that are to be used with
 * the object.
 *
 * This is an opaque data structure.
 */
typedef struct _BaseObjectClass BaseObjectClass;

/**
 * The instance holds the data that is associated with the object.
 * it also holds a pointer to the class of the object.
 *
 * This is an opaque data structure.
 */
typedef struct _BaseObjectInstance BaseObjectInstance;

/* internal structure of the structs are as follows:
typedef struct _BaseObjectClass {
  // BaseObjectInstance * (* initize)(* BaseObjectInstance);
  void (* deinitize)(* BaseObjectInstance);
  int (* ref)(* BaseObjectInstance);
  int (* unref)(* BaseObjectInstance);
  int (* get_refcount)(* BaseObjectInstance);
  char * (* to_string)(* BaseObjectInstance);
};

typedef struct _BaseObjectInstance {
  BaseObjectClass * base_class;
  unsigned int refcount = 0;
};
*/

/* -----------------
 * Helper functions...
 * -----------------
 */
/**
 * This function is used to set the method to initize a new instance of an
 * object.
 *
 * This has no use, rely...(?)
 */
// void base_object_set_init_method (BaseObjectInstance * self, BaseObjectInstance * (* method)(* BaseObjectInstance));

/**
 * This function is used to set the method to deinitize an object.
 * 
 * set it to a method that deinitize your object.
 */
void base_object_set_deinit_method (BaseObjectInstance * self, void (* method)(* BaseObjectInstance));

/**
 * This fuction is used to set the ref method.
 * 
 * DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.
 */
void base_object_set_ref_method (BaseObjectInstance * self, int (* method)(* BaseObjectInstance));

/**
 * This fuction is used to set the unref method.
 * 
 * DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.
 */
void base_object_set_unref_method (BaseObjectInstance * self, int (* method)(* BaseObjectInstance));

/**
 * This fuction is used to set the get_refcount method.
 * 
 * DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.
 */
void base_object_set_get_refcount_method (BaseObjectInstance * self, int (* method)(* BaseObjectInstance));

/**
 * This fuction is used to set the to_string method.
 */
void base_object_set_to_string_method (BaseObjectInstance * self, char * (* method)(* BaseObjectInstance));


/*
 * concreate functions are defined in the C file.
 */

/* ----------------------
 * Base object functions.
 * ----------------------
 */

/**
 * This function initizes an intance of the BaseObject, it also sets the methods
 * to be used with the object and sets the refrence count to one.
 */
void base_object_initize (BaseObjectInstance * self);

/**
 * This function creates a new base object.
 */
BaseObjectInstance * base_object_new ();

/**
 * This function deinitizes/frees an object even if it is still referenced.
 * This is usualy a bad idea, use base_object_unref instead.
 */
void base_object_free (BaseObjectInstance * self);

/**
 * This function gets the class (which hold the object methods).
 */
BaseObjectClass * base_object_get_class (BaseObjectInstance * self);

/**
 * This function sets the instance class of an object.
 */
void base_object_set_class (BaseObjectInstance * self, BaseObjectClass * klass);

/**
 * This function is used to decrese the reference count of an object.
 * When an object reaches zero, it will deinitize the object using the objects
 * deititize method.
 * 
 * It returns the current (after change) reference count.
 */
int base_object_unref (BaseObjectInstance * self);

/**
 * This function is used to increse the reference count of an object.
 *
 * Returns the current (after change) reference count.
 */
int base_object_ref (BaseObjectInstance * self);

/**
 * This function returns the current reference count without chaning it.
 */
int base_object_get_refcount (BaseObjectInstance * self);

/**
 * This function returns a textual (string) that represesnts the object.
 * The method can be set using base_object_set_to_string_method.
 *
 * Note: The string that is returned must be freed.
 */
char * base_object_to_string (BaseObjectInstance * self);


#endif
